"""Tests for core/compaction.py (enhanced compaction)."""
import pytest


def test_estimate_tokens_empty():
    from salmalm.core.compaction import estimate_tokens
    assert estimate_tokens("") == 0


def test_estimate_tokens_english():
    from salmalm.core.compaction import estimate_tokens
    t = estimate_tokens("hello world this is a test")
    assert 5 < t < 15


def test_estimate_tokens_korean():
    from salmalm.core.compaction import estimate_tokens
    t = estimate_tokens("안녕하세요 세계")
    assert t >= 7  # Each Korean char ~1 token


def test_importance_score_system():
    from salmalm.core.compaction import _importance_score
    assert _importance_score({"role": "system", "content": "x"}) == 10.0


def test_importance_score_tool_low():
    from salmalm.core.compaction import _importance_score
    score = _importance_score({"role": "tool", "content": "result data"})
    assert score < 1.0


def test_importance_score_tool_error():
    from salmalm.core.compaction import _importance_score
    score = _importance_score({"role": "tool", "content": "Error: not found"})
    assert score > 0.5


def test_importance_score_user_decision():
    from salmalm.core.compaction import _importance_score
    score = _importance_score({"role": "user", "content": "Always use Korean for comments"})
    assert score >= 3.0


def test_split_by_importance():
    from salmalm.core.compaction import _split_by_importance
    msgs = [
        {"role": "system", "content": "sys"},
        {"role": "user", "content": "old1"},
        {"role": "assistant", "content": "old2"},
        {"role": "user", "content": "recent1"},
        {"role": "assistant", "content": "recent2"},
    ]
    sys, old, recent = _split_by_importance(msgs, keep_recent=2)
    assert len(sys) == 1
    assert len(recent) == 2
    assert len(old) == 2


def test_extract_key_facts():
    from salmalm.core.compaction import _extract_key_facts
    msgs = [
        {"role": "user", "content": "Always remember to use TypeScript for all new code"},
        {"role": "assistant", "content": "short"},
        {"role": "assistant", "content": "```python\ndef foo(): pass\n```" + "x" * 500},
    ]
    facts = _extract_key_facts(msgs)
    assert len(facts) >= 1
    assert any("TypeScript" in f for f in facts)


def test_enhanced_compact_short_noop():
    from salmalm.core.compaction import enhanced_compact
    msgs = [
        {"role": "system", "content": "sys"},
        {"role": "user", "content": "hi"},
        {"role": "assistant", "content": "hello"},
    ]
    result = enhanced_compact(msgs, target_tokens=1000)
    assert len(result) == 3  # No compaction needed


def test_msg_text_string():
    from salmalm.core.compaction import _msg_text
    assert _msg_text({"content": "hello"}) == "hello"


def test_msg_text_list():
    from salmalm.core.compaction import _msg_text
    msg = {"content": [{"type": "text", "text": "hello"}, {"type": "image"}]}
    assert "hello" in _msg_text(msg)



"""Session Groups (대화 주제 그룹) — LobeChat style."""

from __future__ import annotations

from datetime import datetime
from typing import Dict, List, Optional

from salmalm.constants import KST
from salmalm.security.crypto import log


class SessionGroupManager:
    """Manage session groups/folders for organizing conversations."""

    def __init__(self) -> None:
        """Init  ."""
        self._ensure_tables()

    def _ensure_tables(self):
        """Ensure tables."""
        try:
            from salmalm.core import _get_db

            conn = _get_db()
            conn.execute("""CREATE TABLE IF NOT EXISTS session_groups (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                name TEXT NOT NULL UNIQUE,
                color TEXT DEFAULT '#6366f1',
                sort_order INTEGER DEFAULT 0,
                collapsed INTEGER DEFAULT 0,
                created_at TEXT NOT NULL
            )""")
            try:
                conn.execute("ALTER TABLE session_store ADD COLUMN group_id INTEGER DEFAULT NULL")
            except Exception as e:  # noqa: broad-except
                log.debug(f"Suppressed: {e}")
            conn.commit()
            row = conn.execute("SELECT COUNT(*) FROM session_groups").fetchone()
            if row[0] == 0:
                now = datetime.now(KST).isoformat()
                conn.execute(
                    "INSERT INTO session_groups (name, sort_order, created_at) VALUES (?, 0, ?)", ("기본", now)
                )
                conn.commit()
        except Exception as e:
            log.warning(f"Session groups init: {e}")

    def list_groups(self) -> List[Dict]:
        """List groups."""
        try:
            from salmalm.core import _get_db

            conn = _get_db()
            rows = conn.execute(
                "SELECT id, name, color, sort_order, collapsed, created_at FROM session_groups ORDER BY sort_order, id"
            ).fetchall()
            groups = []
            for r in rows:
                count = conn.execute("SELECT COUNT(*) FROM session_store WHERE group_id=?", (r[0],)).fetchone()[0]
                groups.append(
                    {
                        "id": r[0],
                        "name": r[1],
                        "color": r[2],
                        "sort_order": r[3],
                        "collapsed": bool(r[4]),
                        "created_at": r[5],
                        "session_count": count,
                    }
                )
            return groups
        except Exception as e:  # noqa: broad-except
            return []

    def create_group(self, name: str, color: str = "#6366f1") -> Dict:
        """Create group."""
        from salmalm.core import _get_db

        conn = _get_db()
        now = datetime.now(KST).isoformat()
        max_order = conn.execute("SELECT COALESCE(MAX(sort_order),0) FROM session_groups").fetchone()[0]
        conn.execute(
            "INSERT INTO session_groups (name, color, sort_order, created_at) VALUES (?, ?, ?, ?)",
            (name, color, max_order + 1, now),
        )
        conn.commit()
        gid = conn.execute("SELECT last_insert_rowid()").fetchone()[0]
        return {"id": gid, "name": name, "color": color, "ok": True}

    _ALLOWED_UPDATE_COLS = frozenset({"name", "color", "sort_order", "collapsed"})

    def update_group(self, group_id: int, **kwargs) -> bool:
        """Update group."""
        from salmalm.core import _get_db

        for key in kwargs:
            if key not in self._ALLOWED_UPDATE_COLS:
                raise ValueError(f"Invalid column: {key}")

        conn = _get_db()
        sets = []
        vals = []
        for key in sorted(self._ALLOWED_UPDATE_COLS):
            if key in kwargs:
                sets.append(f"{key}=?")
                vals.append(kwargs[key])
        if not sets:
            return False
        vals.append(group_id)
        conn.execute(f"UPDATE session_groups SET {','.join(sets)} WHERE id=?", vals)
        conn.commit()
        return True

    def delete_group(self, group_id: int) -> bool:
        """Delete group."""
        from salmalm.core import _get_db

        conn = _get_db()
        conn.execute("UPDATE session_store SET group_id=NULL WHERE group_id=?", (group_id,))
        conn.execute("DELETE FROM session_groups WHERE id=?", (group_id,))
        conn.commit()
        return True

    def move_session(self, session_id: str, group_id: Optional[int]) -> bool:
        """Move session."""
        from salmalm.core import _get_db

        conn = _get_db()
        conn.execute("UPDATE session_store SET group_id=? WHERE session_id=?", (group_id, session_id))
        conn.commit()
        return True


session_groups = SessionGroupManager()



"""Tests for Workflow Engine."""
import json
import os
import sys
import shutil
import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch

sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from salmalm.features.workflow import (
    WorkflowEngine, _substitute, _substitute_params, _eval_condition,
    handle_workflow_command, WORKFLOWS_DIR
)


class TestVariableSubstitution(unittest.TestCase):

    def test_simple_substitution(self):
        ctx = {'step1': {'result': 'hello', 'count': 5}}
        self.assertEqual(_substitute('{{step1.result}}', ctx), 'hello')

    def test_multiple_substitutions(self):
        ctx = {'a': {'result': 'X'}, 'b': {'result': 'Y'}}
        self.assertEqual(_substitute('{{a.result}} and {{b.result}}', ctx), 'X and Y')

    def test_missing_var(self):
        self.assertEqual(_substitute('{{missing.result}}', {}), '{{missing.result}}')

    def test_substitute_params(self):
        ctx = {'s1': {'result': 'data'}}
        params = {'msg': 'Got: {{s1.result}}', 'count': 5}
        result = _substitute_params(params, ctx)
        self.assertEqual(result['msg'], 'Got: data')
        self.assertEqual(result['count'], 5)


class TestConditionEval(unittest.TestCase):

    def test_greater_than(self):
        self.assertTrue(_eval_condition('5 > 0', {}))

    def test_less_than(self):
        self.assertFalse(_eval_condition('0 > 5', {}))

    def test_equal(self):
        self.assertTrue(_eval_condition('3 == 3', {}))

    def test_with_substitution(self):
        ctx = {'s': {'count': '10'}}
        self.assertTrue(_eval_condition('{{s.count}} > 0', ctx))


class TestWorkflowEngine(unittest.TestCase):

    def setUp(self):
        self.orig_dir = WORKFLOWS_DIR
        self.tmpdir = tempfile.mkdtemp()
        # Patch WORKFLOWS_DIR
        import salmalm.features.workflow as wm
        wm.WORKFLOWS_DIR = Path(self.tmpdir)
        wm.WORKFLOW_LOG_DIR = Path(self.tmpdir) / 'logs'
        self.engine = WorkflowEngine(tool_executor=lambda tool, params: f'result_{tool}')

    def tearDown(self):
        shutil.rmtree(self.tmpdir, ignore_errors=True)
        import salmalm.features.workflow as wm
        wm.WORKFLOWS_DIR = self.orig_dir
        wm.WORKFLOW_LOG_DIR = self.orig_dir / 'logs'

    def test_save_and_get(self):
        wf = {'name': 'test1', 'steps': [{'id': 's1', 'tool': 'echo', 'params': {}}]}
        self.engine.save_workflow(wf)
        loaded = self.engine.get_workflow('test1')
        self.assertEqual(loaded['name'], 'test1')

    def test_list_workflows(self):
        self.engine.save_workflow({'name': 'wf1', 'steps': [], 'trigger': {'type': 'manual'}})
        self.engine.save_workflow({'name': 'wf2', 'steps': [{'id': 's'}], 'trigger': {}})
        wfs = self.engine.list_workflows()
        names = [w['name'] for w in wfs]
        self.assertIn('wf1', names)
        self.assertIn('wf2', names)

    def test_delete_workflow(self):
        self.engine.save_workflow({'name': 'del_me', 'steps': []})
        result = self.engine.delete_workflow('del_me')
        self.assertIn('삭제', result)
        self.assertIsNone(self.engine.get_workflow('del_me'))

    def test_execute_simple(self):
        wf = {
            'name': 'simple',
            'steps': [{'id': 's1', 'tool': 'echo', 'params': {'msg': 'hi'}}],
            'on_error': 'stop',
        }
        result = self.engine.execute(wf)
        self.assertTrue(result['success'])
        self.assertEqual(len(result['results']), 1)

    def test_execute_with_condition_false(self):
        wf = {
            'name': 'cond',
            'steps': [
                {'id': 's1', 'tool': 'echo', 'params': {}},
                {'id': 's2', 'tool': 'echo', 'params': {}, 'if': '{{s1.count}} > 5'},
            ],
        }
        result = self.engine.execute(wf)
        self.assertTrue(any('skipped' in str(r.get('result', '')) for r in result['results']))

    def test_execute_error_stop(self):
        def failing_executor(tool, params):
            raise RuntimeError('fail')
        engine = WorkflowEngine(tool_executor=failing_executor)
        wf = {
            'name': 'fail',
            'steps': [{'id': 's1', 'tool': 'x', 'params': {}}, {'id': 's2', 'tool': 'y', 'params': {}}],
            'on_error': 'stop',
        }
        result = engine.execute(wf)
        self.assertFalse(result['success'])
        # Should stop after first failure
        self.assertEqual(len(result['results']), 1)

    def test_get_presets(self):
        presets = self.engine.get_presets()
        self.assertTrue(len(presets) >= 3)
        names = [p['name'] for p in presets]
        self.assertIn('morning_briefing', names)

    def test_install_preset(self):
        result = self.engine.install_preset('morning_briefing')
        self.assertIn('저장', result)

    def test_get_logs_empty(self):
        logs = self.engine.get_logs('nonexistent')
        self.assertEqual(logs, [])

    def test_run_nonexistent(self):
        result = self.engine.run('no_such_workflow')
        self.assertFalse(result['success'])


class TestWorkflowCommands(unittest.TestCase):

    def test_command_list(self):
        result = handle_workflow_command('/workflow list')
        self.assertIsInstance(result, str)

    def test_command_presets(self):
        result = handle_workflow_command('/workflow presets')
        self.assertIn('프리셋', result)


if __name__ == '__main__':
    unittest.main()



"""salmalm.core.classifier — task classification package.

Re-exports all public symbols from sub-modules for backward compatibility.
Original module: salmalm/core/classifier.py (1325 lines, now split into 3 sub-modules).
"""
from __future__ import annotations

# Intent classification
from salmalm.core.classifier.intent import (  # noqa: F401
    TaskClassifier,
    classify_task,
)

# Keyword/pattern constants
from salmalm.core.classifier.keywords import (  # noqa: F401
    INTENT_TOOLS,
    _KEYWORD_TOOLS,
    _EMOJI_TOOLS,
    _QUESTION_WORDS,
    _QUESTION_INJECT_TOOLS,
    _TIME_PATTERN_RE,
    _TIME_INJECT_TOOLS,
    get_extra_tools,
)

# Token allocation
from salmalm.core.classifier.tokens import (  # noqa: F401
    INTENT_MAX_TOKENS,
    _DETAIL_KEYWORDS,
    _get_dynamic_max_tokens,
)

# Compat aliases
from salmalm.core.cost import estimate_tokens  # noqa: F401
classify_intent = classify_task  # noqa: F401 — backward-compat alias

__all__ = [
    "TaskClassifier",
    "classify_task",
    "classify_intent",
    "INTENT_TOOLS",
    "_KEYWORD_TOOLS",
    "_EMOJI_TOOLS",
    "_QUESTION_WORDS",
    "_QUESTION_INJECT_TOOLS",
    "_TIME_PATTERN_RE",
    "_TIME_INJECT_TOOLS",
    "get_extra_tools",
    "INTENT_MAX_TOKENS",
    "_DETAIL_KEYWORDS",
    "_get_dynamic_max_tokens",
    "estimate_tokens",
]



"""SalmAlm configuration constants."""

from .limits import (
    CACHE_TTL,
    COMPACTION_THRESHOLD,
    DEFAULT_MAX_TOKENS,
    INTENT_COMPLEX_MSG,
    INTENT_CONTEXT_DEPTH,
    INTENT_SHORT_MSG,
    REFLECT_SNIPPET_LEN,
)
from .models import FALLBACK_MODELS, MODEL_ALIASES, MODELS, THINKING_BUDGET_MAP
from .paths import (
    AGENTS_FILE,
    AUDIT_DB,
    BASE_DIR,
    CACHE_DB,
    DATA_DIR,
    LOG_FILE,
    MEMORY_DB,
    MEMORY_DIR,
    MEMORY_FILE,
    SOUL_FILE,
    TOOLS_FILE,
    USER_FILE,
    VAULT_FILE,
    WORKSPACE_DIR,
)



import logging
import sys

try:
    from importlib.metadata import version as _pkg_version

    __version__ = _pkg_version("salmalm")
except Exception:
    __version__ = "0.0.0-dev"  # fallback — metadata missing

log = logging.getLogger("salmalm")
log.addHandler(logging.NullHandler())  # Prevent "No handlers" warning at import
app = None  # Will be set below if runtime (not during build)


def _register_services() -> None:
    """Register all service factories. Each registration is isolated so one
    failure does not prevent the others from loading."""
    if app is None:
        return
    import logging as _log
    _svc_log = _log.getLogger(__name__)
    _services = [
        ("vault", "salmalm.security.crypto", "vault"),
        ("router", "salmalm.core", "router"),
        ("auth_manager", "salmalm.auth", "auth_manager"),
        ("rate_limiter", "salmalm.auth", "rate_limiter"),
        ("rag_engine", "salmalm.rag", "rag_engine"),
        ("mcp_manager", "salmalm.mcp", "mcp_manager"),
        ("node_manager", "salmalm.nodes", "node_manager"),
        ("health_monitor", "salmalm.stability", "health_monitor"),
        ("telegram_bot", "salmalm.telegram", "telegram_bot"),
        ("user_manager", "salmalm.users", "user_manager"),
        ("discord_bot", "salmalm.discord_bot", "discord_bot"),
        ("ws_server", "salmalm.ws", "ws_server"),
    ]
    for svc_name, module_path, attr in _services:
        try:
            app.register(svc_name, lambda m=module_path, a=attr: getattr(__import__(m, fromlist=[a]), a))
        except Exception as _e:
            _svc_log.warning(f"[INIT] Failed to register service '{svc_name}': {_e}")


_logging_initialized = False


def init_logging() -> None:
    """Initialize file + console logging. Called once from entrypoint, not import."""
    global _logging_initialized
    if _logging_initialized:
        return
    # NullHandler doesn't count — it's a library default, not real logging
    if log.handlers and not all(isinstance(h, logging.NullHandler) for h in log.handlers):
        return
    _logging_initialized = True
    try:
        from .constants import LOG_FILE, DATA_DIR

        DATA_DIR.mkdir(parents=True, exist_ok=True)
        log.setLevel(logging.INFO)
        log.addHandler(logging.FileHandler(LOG_FILE, encoding="utf-8"))
        _sh = logging.StreamHandler(sys.stdout)
        _sh.encoding = "utf-8"  # type: ignore[attr-defined]
        log.addHandler(_sh)
        for h in log.handlers:
            h.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s] %(message)s"))
    except Exception as e:  # noqa: broad-except
        log.debug(f"Suppressed: {e}")


import os as _os

_BUILDING = _os.environ.get("SALMALM_BUILDING") == "1"
if not _BUILDING:
    try:
        from .container import Container

        app = Container()
        _register_services()
    except Exception as e:
        log.debug(f"[INIT] Container setup failed (non-fatal at import): {e}")



# Subpackage proxy for backward compatibility
from salmalm.utils.shim import install_shim as _install_shim

_install_shim(
    __name__,
    "salmalm.tools.tools",
    {
        "tools",
        "tool_handlers",
        "tool_registry",
        "tools_agent",
        "tools_brave",
        "tools_browser",
        "tools_calendar",
        "tools_common",
        "tools_email",
        "tools_exec",
        "tools_file",
        "tools_google",
        "tools_media",
        "tools_memory",
        "tools_misc",
        "tools_patch",
        "tools_personal",
        "tools_reaction",
        "tools_reminder",
        "tools_system",
        "tools_util",
        "tools_web",
    },
)

# Re-export commonly imported names for backward compat
from salmalm.tools.tool_handlers import execute_tool  # noqa: F401
try:
    from salmalm.tools.tool_registry import get_all_tools as TOOL_DEFINITIONS  # noqa: F401
except Exception:
    pass



"""SalmAlm Retry Policy — exponential backoff with jitter.

Provides robust retry logic for API calls:
- Exponential backoff: 1s → 2s → 4s (configurable)
- Jitter: ±10% randomization (thundering herd prevention)
- 429 Rate Limit: respects Retry-After header
- 5xx: retries
- 4xx (400, 401, 403): no retry
- Timeout: retries
- 529 Overloaded: 30s wait + retry
"""

from __future__ import annotations

import functools
import random
import time
import urllib.error
from typing import Any, Callable, Optional

from salmalm.security.crypto import log

# ── Constants ──
DEFAULT_MAX_ATTEMPTS = 3
DEFAULT_BASE_DELAY = 1.0
DEFAULT_MAX_DELAY = 30.0
JITTER_FACTOR = 0.10  # ±10%
OVERLOADED_WAIT = 30.0  # 529 wait time

# Non-retryable HTTP status codes
_NO_RETRY_CODES = frozenset({400, 401, 403, 404, 405})


def _should_retry(error: Exception) -> tuple:
    """Determine if an error is retryable.

    Returns (should_retry: bool, wait_seconds: float | None).
    wait_seconds is set for rate-limit/overloaded responses.
    """
    if isinstance(error, urllib.error.HTTPError):
        code = error.code
        if code in _NO_RETRY_CODES:
            return False, None
        if code == 429:
            # Respect Retry-After header
            retry_after = error.headers.get("Retry-After") if hasattr(error, "headers") else None
            if retry_after:
                try:
                    wait = float(retry_after)
                    return True, min(wait, DEFAULT_MAX_DELAY)
                except (ValueError, TypeError):
                    pass
            return True, DEFAULT_BASE_DELAY
        if code == 529:
            return True, OVERLOADED_WAIT
        if 500 <= code < 600:
            return True, None
        return False, None

    if isinstance(error, urllib.error.URLError):
        return True, None  # Network error, retry

    if isinstance(error, (TimeoutError, ConnectionError, OSError)):
        return True, None

    if isinstance(error, ValueError):
        msg = str(error).lower()
        if "rate limit" in msg or "429" in msg:
            return True, DEFAULT_BASE_DELAY
        if "overloaded" in msg or "529" in msg:
            return True, OVERLOADED_WAIT
        if "timeout" in msg:
            return True, None
        return False, None

    return False, None


def _add_jitter(delay: float) -> float:
    """Add ±10% jitter to delay."""
    jitter = delay * JITTER_FACTOR
    return delay + random.uniform(-jitter, jitter)  # not security-sensitive: retry jitter


def retry_with_backoff(
    fn: Optional[Callable] = None,
    *,
    max_attempts: int = DEFAULT_MAX_ATTEMPTS,
    base_delay: float = DEFAULT_BASE_DELAY,
    max_delay: float = DEFAULT_MAX_DELAY,
):
    """Decorator: retry function with exponential backoff + jitter.

    Usage:
        @retry_with_backoff
        def my_api_call():
            ...

        @retry_with_backoff(max_attempts=5, base_delay=2.0)
        def my_api_call():
            ...
    """

    def decorator(func: Callable) -> Callable:
        import asyncio as _asyncio
        if _asyncio.iscoroutinefunction(func):
            raise TypeError(
                f"retry_with_backoff cannot decorate async function '{func.__name__}'. "
                "Use async_retry_with_backoff() instead."
            )

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            """Wrapper."""
            last_error = None
            for attempt in range(1, max_attempts + 1):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    last_error = e
                    should_retry, forced_wait = _should_retry(e)

                    if not should_retry or attempt >= max_attempts:
                        log.warning(f"[RETRY] {func.__name__} failed after {attempt} attempts: {e}")
                        raise

                    if forced_wait is not None:
                        wait = _add_jitter(forced_wait)
                    else:
                        delay = base_delay * (2 ** (attempt - 1))
                        wait = _add_jitter(min(delay, max_delay))

                    log.info(
                        f"[RETRY] {func.__name__} attempt {attempt}/{max_attempts} "
                        f"failed ({type(e).__name__}), retrying in {wait:.1f}s"
                    )
                    time.sleep(wait)

            raise last_error  # Should not reach here

        return wrapper

    if fn is not None:
        return decorator(fn)
    return decorator


async def async_retry_with_backoff(
    coro_fn: Callable,
    *args,
    max_attempts: int = DEFAULT_MAX_ATTEMPTS,
    base_delay: float = DEFAULT_BASE_DELAY,
    max_delay: float = DEFAULT_MAX_DELAY,
    **kwargs,
) -> Any:
    """Async version: retry an async callable with exponential backoff.

    Usage:
        result = await async_retry_with_backoff(my_async_fn, arg1, arg2)
    """
    import asyncio

    last_error = None
    for attempt in range(1, max_attempts + 1):
        try:
            return await coro_fn(*args, **kwargs)
        except Exception as e:
            last_error = e
            should_retry, forced_wait = _should_retry(e)

            if not should_retry or attempt >= max_attempts:
                log.warning(f"[RETRY] {coro_fn.__name__} failed after {attempt} attempts: {e}")
                raise

            if forced_wait is not None:
                wait = _add_jitter(forced_wait)
            else:
                delay = base_delay * (2 ** (attempt - 1))
                wait = _add_jitter(min(delay, max_delay))

            log.info(
                f"[RETRY] {coro_fn.__name__} attempt {attempt}/{max_attempts} "
                f"failed ({type(e).__name__}), retrying in {wait:.1f}s"
            )
            await asyncio.sleep(wait)

    raise last_error


def retry_call(
    fn: Callable, *args, max_attempts: int = DEFAULT_MAX_ATTEMPTS, base_delay: float = DEFAULT_BASE_DELAY, **kwargs
) -> Any:
    """Functional retry: call fn with retry logic (not a decorator).

    Usage:
        result = retry_call(requests.post, url, json=data)
    """
    last_error = None
    for attempt in range(1, max_attempts + 1):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            last_error = e
            should_retry, forced_wait = _should_retry(e)

            if not should_retry or attempt >= max_attempts:
                raise

            if forced_wait is not None:
                wait = _add_jitter(forced_wait)
            else:
                delay = base_delay * (2 ** (attempt - 1))
                wait = _add_jitter(min(delay, DEFAULT_MAX_DELAY))

            log.info(f"[RETRY] attempt {attempt}/{max_attempts}: {type(e).__name__}, wait {wait:.1f}s")
            time.sleep(wait)

    raise last_error



"""Tests for graceful shutdown, health endpoint, plugin hot-reload, and conversation export."""
import asyncio
import json
import os
import sys
import tempfile
import threading
import time
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

# Ensure project root on path
sys.path.insert(0, str(Path(__file__).parent.parent))


class TestHealthEndpoint(unittest.TestCase):
    """Test health report generation."""

    def test_get_health_report_returns_dict(self):
        from salmalm.core.health import get_health_report
        report = get_health_report()
        self.assertIsInstance(report, dict)
        self.assertIn('status', report)
        self.assertIn('version', report)
        self.assertIn('uptime_seconds', report)
        self.assertIn('uptime_human', report)
        self.assertIn('memory_mb', report)
        self.assertIn('active_sessions', report)
        self.assertIn('disk', report)
        self.assertIn('threads', report)

    def test_health_status_values(self):
        from salmalm.core.health import get_health_report
        report = get_health_report()
        self.assertIn(report['status'], ('healthy', 'degraded', 'unhealthy'))

    def test_uptime_is_positive(self):
        from salmalm.core.health import get_health_report
        report = get_health_report()
        self.assertGreaterEqual(report['uptime_seconds'], 0)

    def test_memory_mb_is_number(self):
        from salmalm.core.health import get_health_report
        report = get_health_report()
        self.assertIsInstance(report['memory_mb'], (int, float))

    def test_disk_info(self):
        from salmalm.core.health import get_health_report
        report = get_health_report()
        disk = report.get('disk', {})
        # On Linux, should have free_mb and total_mb
        if disk:
            self.assertIn('free_mb', disk)
            self.assertIn('total_mb', disk)

    def test_llm_status_structure(self):
        from salmalm.core.health import get_health_report
        report = get_health_report()
        llm = report.get('llm', {})
        self.assertIn('connected', llm)

    def test_format_uptime(self):
        from salmalm.core.health import _format_uptime
        result = _format_uptime()
        self.assertIsInstance(result, str)
        # Should contain 'h' and 'm'
        self.assertTrue('h' in result or 'd' in result or 'm' in result)


class TestConversationExport(unittest.TestCase):
    """Test conversation export to md/json/html."""

    def _make_session(self):
        """Create a mock session for testing."""
        session = MagicMock()
        session.id = 'test-session-123'
        session.messages = [
            {'role': 'system', 'content': 'You are a helpful assistant.'},
            {'role': 'user', 'content': 'Hello, how are you?'},
            {'role': 'assistant', 'content': 'I am doing well, thank you!'},
            {'role': 'user', 'content': 'Tell me about Python.'},
            {'role': 'assistant', 'content': 'Python is a great programming language.'},
        ]
        return session

    def test_export_markdown(self):
        from salmalm.core.export import export_session
        session = self._make_session()
        result = export_session(session, fmt='md')
        self.assertTrue(result['ok'])
        self.assertTrue(result['filename'].endswith('.md'))
        self.assertGreater(result['size'], 0)
        # Check file content
        content = Path(result['path']).read_text(encoding='utf-8')
        self.assertIn('# SalmAlm Conversation Export', content)
        self.assertIn('Hello, how are you?', content)
        self.assertIn('Python is a great programming language.', content)
        # System message should be excluded
        self.assertNotIn('You are a helpful assistant.', content)
        # Cleanup
        os.unlink(result['path'])

    def test_export_json(self):
        from salmalm.core.export import export_session
        session = self._make_session()
        result = export_session(session, fmt='json')
        self.assertTrue(result['ok'])
        self.assertTrue(result['filename'].endswith('.json'))
        content = Path(result['path']).read_text(encoding='utf-8')
        data = json.loads(content)
        self.assertEqual(data['session_id'], 'test-session-123')
        self.assertGreater(data['message_count'], 0)
        # System messages excluded
        roles = [m['role'] for m in data['messages']]
        self.assertNotIn('system', roles)
        os.unlink(result['path'])

    def test_export_html(self):
        from salmalm.core.export import export_session
        session = self._make_session()
        result = export_session(session, fmt='html')
        self.assertTrue(result['ok'])
        self.assertTrue(result['filename'].endswith('.html'))
        content = Path(result['path']).read_text(encoding='utf-8')
        self.assertIn('<!DOCTYPE html>', content)
        self.assertIn('Hello, how are you?', content)
        os.unlink(result['path'])

    def test_export_invalid_format(self):
        from salmalm.core.export import export_session
        session = self._make_session()
        result = export_session(session, fmt='pdf')
        self.assertFalse(result['ok'])
        self.assertIn('Unsupported format', result['error'])

    def test_export_multimodal_content(self):
        """Test export handles list-type content blocks."""
        from salmalm.core.export import export_session
        session = self._make_session()
        session.messages.append({
            'role': 'user',
            'content': [
                {'type': 'text', 'text': 'Analyze this image'},
                {'type': 'image', 'source': {'type': 'base64', 'data': 'abc'}},
            ]
        })
        result = export_session(session, fmt='md')
        self.assertTrue(result['ok'])
        content = Path(result['path']).read_text(encoding='utf-8')
        self.assertIn('Analyze this image', content)
        os.unlink(result['path'])

    def test_export_tool_result_content(self):
        """Test export handles tool_result blocks."""
        from salmalm.core.export import export_session
        session = self._make_session()
        session.messages.append({
            'role': 'user',
            'content': [
                {'type': 'tool_result', 'tool_use_id': 'abc', 'content': 'tool output here'},
            ]
        })
        result = export_session(session, fmt='json')
        self.assertTrue(result['ok'])
        os.unlink(result['path'])


class TestExportCommand(unittest.TestCase):
    """Test /export command handler."""

    def test_cmd_export_default_md(self):
        from salmalm.core.engine import _cmd_export
        session = MagicMock()
        session.id = 'cmd-test'
        session.messages = [
            {'role': 'user', 'content': 'test'},
            {'role': 'assistant', 'content': 'reply'},
        ]
        result = _cmd_export('/export', session)
        self.assertIn('exported', result)
        self.assertIn('MD', result)

    def test_cmd_export_json(self):
        from salmalm.core.engine import _cmd_export
        session = MagicMock()
        session.id = 'cmd-test-json'
        session.messages = [
            {'role': 'user', 'content': 'test'},
            {'role': 'assistant', 'content': 'reply'},
        ]
        result = _cmd_export('/export json', session)
        self.assertIn('exported', result)
        self.assertIn('JSON', result)

    def test_cmd_export_html(self):
        from salmalm.core.engine import _cmd_export
        session = MagicMock()
        session.id = 'cmd-test-html'
        session.messages = [
            {'role': 'user', 'content': 'test'},
            {'role': 'assistant', 'content': 'reply'},
        ]
        result = _cmd_export('/export html', session)
        self.assertIn('exported', result)
        self.assertIn('HTML', result)


class TestShutdownManager(unittest.TestCase):
    """Test graceful shutdown manager."""

    def test_shutdown_manager_exists(self):
        from salmalm.core.shutdown import shutdown_manager
        self.assertFalse(shutdown_manager.is_shutting_down)

    def test_shutdown_sets_flag(self):
        from salmalm.core.shutdown import ShutdownManager
        sm = ShutdownManager()
        self.assertFalse(sm.is_shutting_down)
        # Can't easily test full async execute without mocking everything,
        # but we can verify the flag mechanism
        sm._shutting_down = True
        self.assertTrue(sm.is_shutting_down)


class TestPluginWatcher(unittest.TestCase):
    """Test plugin hot-reload watcher."""

    def test_plugin_watcher_exists(self):
        from salmalm.core.plugin_watcher import plugin_watcher
        self.assertFalse(plugin_watcher.running)

    def test_extract_plugin_name(self):
        from salmalm.core.plugin_watcher import PluginWatcher, PLUGINS_DIR
        pw = PluginWatcher()
        # Test with a path inside plugins dir
        test_path = PLUGINS_DIR / 'my_plugin' / '__init__.py'
        name = pw._extract_plugin_name(test_path)
        self.assertEqual(name, 'my_plugin')

    def test_extract_plugin_name_outside(self):
        from salmalm.core.plugin_watcher import PluginWatcher
        pw = PluginWatcher()
        name = pw._extract_plugin_name(Path('/tmp/random/file.py'))
        self.assertIsNone(name)

    def test_reload_all(self):
        from salmalm.core.plugin_watcher import PluginWatcher
        pw = PluginWatcher()
        result = pw.reload_all()
        self.assertIn('Reloaded', result)

    def test_start_stop(self):
        from salmalm.core.plugin_watcher import PluginWatcher
        pw = PluginWatcher(interval=1)
        pw.start()
        self.assertTrue(pw.running)
        time.sleep(0.1)
        pw.stop()
        self.assertFalse(pw.running)


class TestPluginsCommand(unittest.TestCase):
    """Test /plugins reload command."""

    def test_plugins_reload(self):
        from salmalm.core.engine import _cmd_plugins
        session = MagicMock()
        result = _cmd_plugins('/plugins reload', session)
        self.assertIn('Reloaded', result)

    def test_plugins_list(self):
        from salmalm.core.engine import _cmd_plugins
        session = MagicMock()
        result = _cmd_plugins('/plugins list', session)
        # Should return either plugins or "no plugins" message
        self.assertIsInstance(result, str)


if __name__ == '__main__':
    unittest.main()



"""Model registry and aliases."""

from __future__ import annotations

MODELS = {
    "opus": "anthropic/claude-opus-4-6",
    "sonnet": "anthropic/claude-sonnet-4-6",
    "haiku": "anthropic/claude-haiku-4-5-20251001",
    "gpt5.2": "openai/gpt-5.2-codex",
    "gpt5.1": "openai/gpt-5.1-codex",
    "gpt4.1": "openai/gpt-4.1",
    "gpt4.1mini": "openai/gpt-4.1-mini",
    "gpt4.1nano": "openai/gpt-4.1-nano",
    "o3": "openai/o3",
    "o4mini": "openai/o4-mini",
    "grok4": "xai/grok-4",
    "grok3": "xai/grok-3",
    "grok3mini": "xai/grok-3-mini",
    "gemini3pro": "google/gemini-3-pro-preview",
    "gemini3flash": "google/gemini-3-flash-preview",
    "gemini2.5pro": "google/gemini-2.5-pro",
    "gemini2.5flash": "google/gemini-2.5-flash",
    "deepseek-r1": "openrouter/deepseek/deepseek-r1",
    "deepseek-chat": "openrouter/deepseek/deepseek-chat",
    "maverick": "openrouter/meta-llama/llama-4-maverick",
    "scout": "openrouter/meta-llama/llama-4-scout",
}

FALLBACK_MODELS = {
    "anthropic": "claude-sonnet-4-20250514",
    "xai": "grok-4",
    "google": "gemini-3-flash-preview",
}

MODEL_ALIASES = {
    "claude": MODELS["sonnet"],
    "sonnet": MODELS["sonnet"],
    "opus": MODELS["opus"],
    "haiku": MODELS["haiku"],
    "gpt": MODELS["gpt5.2"],
    "gpt5": MODELS["gpt5.2"],
    "gpt5.1": MODELS["gpt5.1"],
    "gpt4.1": MODELS["gpt4.1"],
    "4.1mini": MODELS["gpt4.1mini"],
    "4.1nano": MODELS["gpt4.1nano"],
    "o3": MODELS["o3"],
    "o4mini": MODELS["o4mini"],
    "grok": MODELS["grok4"],
    "grok4": MODELS["grok4"],
    "grok3": MODELS["grok3"],
    "grok3mini": MODELS["grok3mini"],
    "gemini": MODELS["gemini3pro"],
    "flash": MODELS["gemini3flash"],
    "deepseek": MODELS["deepseek-r1"],
    "maverick": MODELS["maverick"],
    "scout": MODELS["scout"],
    "llama": "ollama/llama3.3",
    "llama3.2": "ollama/llama3.2",
    "llama3.3": "ollama/llama3.3",
    "qwen": "ollama/qwen3",
    "qwen3": "ollama/qwen3",
}

THINKING_BUDGET_MAP: dict = {
    "low": 4_000,
    "medium": 10_000,
    "high": 16_000,
    "xhigh": 32_000,
}



"""Web Content routes mixin."""

import json
import logging
import re

from salmalm.constants import DATA_DIR, VERSION, WORKSPACE_DIR, MEMORY_DIR, BASE_DIR, AUDIT_DB  # noqa: F401
from salmalm.core.core import get_usage_report  # noqa: F401
from salmalm.security.crypto import vault, log  # noqa: F401

log = logging.getLogger(__name__)

_RE_SESSION_SUMMARY = re.compile(r"^/api/sessions/([^/]+)/summary")
_RE_SESSION_ALT = re.compile(r"^/api/sessions/([^/]+)/alternatives/(\d+)")


class ContentMixin:
    GET_ROUTES = {
        "/api/groups": "_get_groups",
        "/api/soul": "_get_soul",
        "/api/memory/files": "_get_memory_files",
        "/api/mcp": "_get_mcp",
        "/api/rag": "_get_rag",
        "/api/personas": "_get_personas",
        "/api/thoughts": "_get_thoughts",
        "/api/thoughts/stats": "_get_thoughts_stats",
        "/api/tools/list": "_get_tools_list",
        "/api/browser/status": "_get_api_browser_status",
        "/api/commands": "_get_commands",
        "/api/dashboard": "_get_api_dashboard",
        "/manifest.json": "_get_manifest_json",
    }
    POST_ROUTES = {
        "/api/bookmarks": "_post_api_bookmarks",
        "/api/thoughts": "_post_api_thoughts",
    }
    GET_PREFIX_ROUTES = [
        ("/api/search", "_get_api_search", None),
        ("/api/rag/search", "_get_api_rag_search", None),
        ("/api/sessions/", "_get_api_sessions_summary", "/summary"),
        ("/api/sessions/", "_get_api_sessions_alternatives", "/alternatives"),
        ("/api/memory/read?", "_get_api_memory_read", None),
    ]

    def _get_thoughts(self):
        """Get thoughts."""
        from salmalm.features.thoughts import thought_stream
        import urllib.parse as _up

        qs = _up.parse_qs(_up.urlparse(self.path).query)
        search_q = qs.get("q", [""])[0]
        if search_q:
            results = thought_stream.search(search_q)
        else:
            n = int(qs.get("limit", ["20"])[0])
            results = thought_stream.list_recent(n)
        self._json({"thoughts": results})

    def _get_thoughts_stats(self):
        """Get thoughts stats."""
        from salmalm.features.thoughts import thought_stream

        self._json(thought_stream.stats())

    def _get_tools_list(self):
        """Get tools list."""
        tools = []
        try:
            from salmalm.tools.tool_registry import _HANDLERS, _ensure_modules

            _ensure_modules()
            for name in sorted(_HANDLERS.keys()):
                tools.append({"name": name, "description": ""})
        except Exception as e:
            log.debug(f"Suppressed: {e}")
        if not tools:
            # Fallback: list all known tools from INTENT_TOOLS
            try:
                from salmalm.core.engine import INTENT_TOOLS

                seen = set()
                for cat_tools in INTENT_TOOLS.values():
                    for t in cat_tools:
                        n = t.get("function", {}).get("name", "")
                        if n and n not in seen:
                            seen.add(n)
                            tools.append(
                                {
                                    "name": n,
                                    "description": t.get("function", {}).get("description", ""),
                                }
                            )
            except Exception as e:  # noqa: broad-except
                tools = [
                    {"name": "web_search", "description": "Search the web"},
                    {"name": "bash", "description": "Execute shell commands"},
                    {"name": "file_read", "description": "Read files"},
                    {"name": "file_write", "description": "Write files"},
                    {"name": "browser", "description": "Browser automation"},
                ]
        self._json({"tools": tools, "count": len(tools)})

    def _post_api_thoughts_search(self):
        """Handle /api/thoughts/search."""
        body = self._body
        if not self._require_auth("user"):
            return
        from salmalm.features.thoughts import thought_stream

        q = body.get("q", body.get("query", ""))
        if not q:
            self._json({"error": "query required"}, 400)
            return
        results = thought_stream.search(q)
        self._json({"thoughts": results})

    def _post_api_bookmarks(self):
        """Post api bookmarks."""
        body = self._body
        # Add/remove bookmark — LobeChat style (북마크 추가/제거)
        if not self._require_auth("user"):
            return
        action = body.get("action", "add")
        session_id = body.get("session_id", "")
        message_index = body.get("message_index")
        if not session_id or message_index is None:
            self._json({"error": "Missing session_id or message_index"}, 400)
            return
        from salmalm.features.edge_cases import bookmark_manager

        if action == "add":
            ok = bookmark_manager.add(
                session_id,
                int(message_index),
                content_preview=body.get("preview", ""),
                note=body.get("note", ""),
                role=body.get("role", "assistant"),
            )
            self._json({"ok": ok})
        elif action == "remove":
            ok = bookmark_manager.remove(session_id, int(message_index))
            self._json({"ok": ok})
        else:
            self._json({"error": "Unknown action"}, 400)
        return

    def _post_api_thoughts(self):
        """Post api thoughts."""
        body = self._body
        from salmalm.features.thoughts import thought_stream

        content = body.get("content", "").strip()
        if not content:
            self._json({"error": "content required"}, 400)
            return
        mood = body.get("mood", "neutral")
        tid = thought_stream.add(content, mood=mood)
        self._json({"ok": True, "id": tid})

    def _get_api_browser_status(self):
        """GET /api/browser/status — check if Playwright (browser automation) is available."""
        available = False
        reason = "playwright not installed"
        try:
            import importlib.util
            if importlib.util.find_spec("playwright") is not None:
                # Also verify chromium executable exists
                try:
                    from playwright.sync_api import sync_playwright
                    with sync_playwright() as p:
                        exe = p.chromium.executable_path
                        import os
                        available = os.path.isfile(exe)
                        reason = "chromium found" if available else f"chromium missing: {exe}"
                except Exception as _e:
                    available = False
                    reason = f"playwright import error: {_e}"
            else:
                reason = "playwright package not installed (pip install salmalm[browser])"
        except Exception as _e:
            reason = str(_e)
        self._json({"available": available, "reason": reason})

    def _get_personas(self):
        """Get personas."""
        from salmalm.core.prompt import list_personas, get_active_persona

        session_id = self.headers.get("X-Session-Id", "web")
        personas = list_personas()
        active = get_active_persona(session_id)
        self._json({"personas": personas, "active": active})

    def _get_memory_files(self):
        """Get memory files."""
        if not self._require_auth("user"):
            return
        mem_dir = BASE_DIR / "memory"
        files = []
        # Main memory file
        main_mem = DATA_DIR / "memory.json"
        if main_mem.exists():
            files.append(
                {
                    "name": "memory.json",
                    "size": main_mem.stat().st_size,
                    "path": "memory.json",
                }
            )
        # Memory directory files
        if mem_dir.exists():
            for f in sorted(mem_dir.iterdir(), reverse=True):
                if f.is_file() and f.suffix in (".json", ".md", ".txt"):
                    files.append(
                        {
                            "name": f.name,
                            "size": f.stat().st_size,
                            "path": f"memory/{f.name}",
                        }
                    )
        # Soul file
        soul = DATA_DIR / "soul.md"
        if soul.exists():
            files.append({"name": "soul.md", "size": soul.stat().st_size, "path": "soul.md"})
        self._json({"files": files})

    def _get_mcp(self):
        """Get mcp."""
        if not self._require_auth("user"):
            return
        from salmalm.features.mcp import mcp_manager

        servers = mcp_manager.list_servers()
        all_tools = mcp_manager.get_all_tools()
        self._json({"servers": servers, "total_tools": len(all_tools)})

    def _get_rag(self):
        """Get rag."""
        if not self._require_auth("user"):
            return
        from salmalm.features.rag import rag_engine

        self._json(rag_engine.get_stats())

    def _get_api_search(self) -> None:
        """Handle GET /api/search routes."""
        if not self._require_auth("user"):
            return
        import urllib.parse

        parsed = urllib.parse.urlparse(self.path)
        params = urllib.parse.parse_qs(parsed.query)
        query = params.get("q", [""])[0]
        if not query:
            self._json({"error": "Missing q parameter"}, 400)
            return
        try:
            lim = max(1, min(int(params.get("limit", ["20"])[0]), 200))
        except (TypeError, ValueError, IndexError):
            lim = 20
        from salmalm.core import search_messages

        results = search_messages(query, limit=lim)
        self._json({"query": query, "results": results, "count": len(results)})

    def _get_api_rag_search(self) -> None:
        """Handle GET /api/rag/search routes."""
        if not self._require_auth("user"):
            return
        from salmalm.features.rag import rag_engine
        import urllib.parse

        parsed = urllib.parse.urlparse(self.path)
        params = urllib.parse.parse_qs(parsed.query)
        query = params.get("q", [""])[0]
        if not query:
            self._json({"error": "Missing q parameter"}, 400)
        else:
            try:
                _rag_n = max(1, min(int(params.get("n", ["5"])[0]), 50))
            except (TypeError, ValueError, IndexError):
                _rag_n = 5
            results = rag_engine.search(query, max_results=_rag_n)
            self._json({"query": query, "results": results})

    def _get_api_memory_read(self) -> None:
        """Handle GET /api/memory/read? routes."""
        if not self._require_auth("user"):
            return
        import urllib.parse as _up

        qs = _up.parse_qs(_up.urlparse(self.path).query)
        fpath = qs.get("file", [""])[0]
        if not fpath or ".." in fpath:
            self._json({"error": "Invalid path"}, 400)
            return
        # P0-1: Block absolute paths and resolve to prevent path traversal
        from pathlib import PurePosixPath

        if PurePosixPath(fpath).is_absolute() or "\\" in fpath:
            self._json({"error": "Invalid path"}, 400)
            return
        full = (BASE_DIR / fpath).resolve()
        if not full.is_relative_to(BASE_DIR.resolve()):
            self._json({"error": "Path outside allowed directory"}, 403)
            return
        if not full.exists() or not full.is_file():
            self._json({"error": "File not found"}, 404)
            return
        try:
            content = full.read_text(encoding="utf-8")[:50000]
            self._json({"file": fpath, "content": content, "size": full.stat().st_size})
        except Exception as e:
            log.warning("[FILE] Read error for %s: %s", fpath, e)
            self._json({"error": "File read error — check server logs"}, 500)

    def _get_api_sessions_summary(self) -> None:
        """Handle GET /api/sessions/ routes."""
        # Conversation summary card — BIG-AGI style (대화 요약 카드)
        if not self._require_auth("user"):
            return
        m = _RE_SESSION_SUMMARY.match(self.path)
        if m:
            from salmalm.features.edge_cases import get_summary_card

            card = get_summary_card(m.group(1))
            self._json({"summary": card})
        else:
            self._json({"error": "Invalid path"}, 400)

    def _get_api_sessions_alternatives(self) -> None:
        """Handle GET /api/sessions/ routes."""
        # Conversation fork alternatives — LibreChat style (대화 포크)
        if not self._require_auth("user"):
            return
        m = _RE_SESSION_ALT.match(self.path)
        if m:
            from salmalm.features.edge_cases import conversation_fork

            alts = conversation_fork.get_alternatives(m.group(1), int(m.group(2)))
            self._json({"alternatives": alts})
        else:
            self._json({"error": "Invalid path"}, 400)

    def _get_groups(self):
        """Get groups."""
        if not self._require_auth("user"):
            return
        from salmalm.features.edge_cases import session_groups

        self._json({"groups": session_groups.list_groups()})

    def _get_soul(self):
        """Get soul."""
        if not self._require_auth("user"):
            return
        from salmalm.core.prompt import get_user_soul, USER_SOUL_FILE

        self._json({"content": get_user_soul(), "path": str(USER_SOUL_FILE)})

    def _get_commands(self):
        """Get commands."""
        cmds = [
            {"name": "/help", "desc": "Show help"},
            {"name": "/status", "desc": "Session status"},
            {"name": "/model", "desc": "Switch model"},
            {"name": "/compact", "desc": "Compress context"},
            {"name": "/context", "desc": "Token breakdown"},
            {"name": "/usage", "desc": "Token/cost tracking"},
            {"name": "/think", "desc": "Record thought / thinking level"},
            {"name": "/persona", "desc": "Switch persona"},
            {"name": "/branch", "desc": "Branch conversation"},
            {"name": "/rollback", "desc": "Rollback messages"},
            {"name": "/life", "desc": "Life dashboard"},
            {"name": "/remind", "desc": "Set reminder"},
            {"name": "/expense", "desc": "Track expense"},
            {"name": "/pomodoro", "desc": "Pomodoro timer"},
            {"name": "/note", "desc": "Save note"},
            {"name": "/link", "desc": "Save link"},
            {"name": "/routine", "desc": "Manage routines"},
            {"name": "/shadow", "desc": "Shadow mode"},
            {"name": "/vault", "desc": "Encrypted vault"},
            {"name": "/capsule", "desc": "Time capsule"},
            {"name": "/deadman", "desc": "Dead man's switch"},
            {"name": "/a2a", "desc": "Agent-to-agent"},
            {"name": "/workflow", "desc": "Workflow engine"},
            {"name": "/mcp", "desc": "MCP management"},
            {"name": "/subagents", "desc": "Sub-agents"},
            {"name": "/oauth", "desc": "OAuth setup"},
            {"name": "/bash", "desc": "Run shell command"},
            {"name": "/screen", "desc": "Browser control"},
            {"name": "/evolve", "desc": "Evolving prompt"},
            {"name": "/mood", "desc": "Mood detection"},
            {"name": "/split", "desc": "A/B split response"},
        ]
        self._json({"commands": cmds, "count": len(cmds)})

    def _get_api_dashboard(self):
        """Get api dashboard."""
        if not self._require_auth("user"):
            return
        # Dashboard data: sessions, costs, tools, cron jobs
        from salmalm.core import _sessions, _llm_cron, PluginLoader, SubAgent  # type: ignore[attr-defined]

        sessions_info = [
            {
                "id": s.id,
                "messages": len(s.messages),
                "last_active": s.last_active,
                "created": s.created,
            }
            for s in _sessions.values()
        ]
        cron_jobs = _llm_cron.list_jobs() if _llm_cron else []
        plugins = [{"name": n, "tools": len(p["tools"])} for n, p in PluginLoader._plugins.items()]
        subagents = SubAgent.list_agents()
        usage = get_usage_report()
        # Cost by hour (from audit)
        cost_timeline = []
        try:
            import sqlite3 as _sq

            _conn = _sq.connect(str(AUDIT_DB))  # noqa: F405
            cur = _conn.execute(
                "SELECT substr(ts,1,13) as hour, COUNT(*) as cnt "
                "FROM audit_log WHERE event='tool_exec' "
                "GROUP BY hour ORDER BY hour DESC LIMIT 24"
            )
            cost_timeline = [{"hour": r[0], "count": r[1]} for r in cur.fetchall()]
            _conn.close()
        except Exception as e:
            log.debug(f"Suppressed: {e}")
        self._json(
            {
                "sessions": sessions_info,
                "usage": usage,
                "cron_jobs": cron_jobs,
                "plugins": plugins,
                "subagents": subagents,
                "cost_timeline": cost_timeline,
            }
        )

    def _get_manifest_json(self):
        """Get manifest json."""
        manifest = {
            "name": "SalmAlm — Personal AI Gateway",
            "short_name": "SalmAlm",
            "description": "Your personal AI gateway. 66 tools, 6 providers, 3 core dependencies (uvicorn, fastapi, websockets).",
            "start_url": "/",
            "display": "standalone",
            "background_color": "#0b0d14",
            "theme_color": "#6366f1",
            "orientation": "any",
            "categories": ["productivity", "utilities"],
            "icons": [
                {
                    "src": "/icon-192.svg",
                    "sizes": "192x192",
                    "type": "image/svg+xml",
                    "purpose": "any",
                },
                {
                    "src": "/icon-512.svg",
                    "sizes": "512x512",
                    "type": "image/svg+xml",
                    "purpose": "any maskable",
                },
            ],
        }
        self.send_response(200)
        self._cors()
        self.send_header("Content-Type", "application/manifest+json")
        self.send_header("Cache-Control", "public, max-age=86400")
        self.end_headers()
        self.wfile.write(json.dumps(manifest).encode())


# ── FastAPI router ────────────────────────────────────────────────────────────
import asyncio as _asyncio
from fastapi import APIRouter as _APIRouter, Request as _Request, Depends as _Depends, Query as _Query
from fastapi.responses import JSONResponse as _JSON, Response as _Response, HTMLResponse as _HTML, StreamingResponse as _SR, RedirectResponse as _RR
from salmalm.web.fastapi_deps import require_auth as _auth, optional_auth as _optauth

router = _APIRouter()

@router.get("/api/groups")
async def get_groups(_u=_Depends(_auth)):
    from salmalm.features.edge_cases import session_groups
    return _JSON(content={"groups": session_groups.list_groups()})

@router.get("/api/soul")
async def get_soul(_u=_Depends(_auth)):
    from salmalm.core.prompt import get_user_soul, USER_SOUL_FILE
    return _JSON(content={"content": get_user_soul(), "path": str(USER_SOUL_FILE)})

@router.get("/api/memory/files")
async def get_memory_files(_u=_Depends(_auth)):
    from salmalm.constants import DATA_DIR, BASE_DIR
    mem_dir = BASE_DIR / "memory"
    files = []
    main_mem = DATA_DIR / "memory.json"
    if main_mem.exists():
        files.append({"name": "memory.json", "size": main_mem.stat().st_size, "path": "memory.json"})
    if mem_dir.exists():
        for f in sorted(mem_dir.iterdir(), reverse=True):
            if f.is_file() and f.suffix in (".json", ".md", ".txt"):
                files.append({"name": f.name, "size": f.stat().st_size, "path": f"memory/{f.name}"})
    soul = DATA_DIR / "soul.md"
    if soul.exists():
        files.append({"name": "soul.md", "size": soul.stat().st_size, "path": "soul.md"})
    return _JSON(content={"files": files})

@router.get("/api/mcp")
async def get_mcp(_u=_Depends(_auth)):
    from salmalm.features.mcp import mcp_manager
    servers = mcp_manager.list_servers()
    all_tools = mcp_manager.get_all_tools()
    return _JSON(content={"servers": servers, "total_tools": len(all_tools)})

@router.get("/api/rag")
async def get_rag(_u=_Depends(_auth)):
    from salmalm.features.rag import rag_engine
    return _JSON(content=rag_engine.get_stats())

@router.get("/api/personas")
async def get_personas(request: _Request):
    from salmalm.core.prompt import list_personas, get_active_persona
    session_id = request.headers.get("x-session-id", "web")
    personas = list_personas()
    active = get_active_persona(session_id)
    return _JSON(content={"personas": personas, "active": active})

@router.get("/api/thoughts")
async def get_thoughts(q: str = _Query(None), limit: int = _Query(20), _u=_Depends(_auth)):
    from salmalm.features.thoughts import thought_stream
    if q:
        results = thought_stream.search(q)
    else:
        results = thought_stream.list_recent(limit)
    return _JSON(content={"thoughts": results})

@router.get("/api/thoughts/stats")
async def get_thoughts_stats(_u=_Depends(_auth)):
    from salmalm.features.thoughts import thought_stream
    return _JSON(content=thought_stream.stats())

@router.get("/api/tools/list")
async def get_tools_list(_u=_Depends(_auth)):
    tools = []
    try:
        from salmalm.tools.tool_registry import _HANDLERS, _ensure_modules
        _ensure_modules()
        for name in sorted(_HANDLERS.keys()):
            tools.append({"name": name, "description": ""})
    except Exception:
        pass
    if not tools:
        try:
            from salmalm.core.engine import INTENT_TOOLS
            seen = set()
            for cat_tools in INTENT_TOOLS.values():
                for t in cat_tools:
                    n = t.get("function", {}).get("name", "")
                    if n and n not in seen:
                        seen.add(n)
                        tools.append({"name": n, "description": t.get("function", {}).get("description", "")})
        except Exception:
            tools = [{"name": "web_search", "description": "Search the web"}, {"name": "bash", "description": "Execute shell commands"},
                     {"name": "file_read", "description": "Read files"}, {"name": "file_write", "description": "Write files"},
                     {"name": "browser", "description": "Browser automation"}]
    return _JSON(content={"tools": tools, "count": len(tools)})

@router.get("/api/browser/status")
async def get_browser_status():
    available = False
    reason = "playwright not installed"
    try:
        import importlib.util
        if importlib.util.find_spec("playwright") is not None:
            try:
                from playwright.sync_api import sync_playwright
                with sync_playwright() as p:
                    import os
                    exe = p.chromium.executable_path
                    available = os.path.isfile(exe)
                    reason = "chromium found" if available else f"chromium missing: {exe}"
            except Exception as _e:
                reason = f"playwright import error: {_e}"
        else:
            reason = "playwright package not installed (pip install salmalm[browser])"
    except Exception as _e:
        reason = str(_e)
    return _JSON(content={"available": available, "reason": reason})

@router.get("/api/commands")
async def get_commands():
    cmds = [
        {"name": "/help", "desc": "Show help"}, {"name": "/status", "desc": "Session status"},
        {"name": "/model", "desc": "Switch model"}, {"name": "/compact", "desc": "Compress context"},
        {"name": "/context", "desc": "Token breakdown"}, {"name": "/usage", "desc": "Token/cost tracking"},
        {"name": "/think", "desc": "Record thought / thinking level"}, {"name": "/persona", "desc": "Switch persona"},
        {"name": "/branch", "desc": "Branch conversation"}, {"name": "/rollback", "desc": "Rollback messages"},
        {"name": "/life", "desc": "Life dashboard"}, {"name": "/remind", "desc": "Set reminder"},
        {"name": "/expense", "desc": "Track expense"}, {"name": "/pomodoro", "desc": "Pomodoro timer"},
        {"name": "/note", "desc": "Save note"}, {"name": "/link", "desc": "Save link"},
        {"name": "/routine", "desc": "Manage routines"}, {"name": "/shadow", "desc": "Shadow mode"},
        {"name": "/vault", "desc": "Encrypted vault"}, {"name": "/capsule", "desc": "Time capsule"},
        {"name": "/deadman", "desc": "Dead man's switch"}, {"name": "/a2a", "desc": "Agent-to-agent"},
        {"name": "/workflow", "desc": "Workflow engine"}, {"name": "/mcp", "desc": "MCP management"},
        {"name": "/subagents", "desc": "Sub-agents"}, {"name": "/oauth", "desc": "OAuth setup"},
        {"name": "/bash", "desc": "Run shell command"}, {"name": "/screen", "desc": "Browser control"},
        {"name": "/evolve", "desc": "Evolving prompt"}, {"name": "/mood", "desc": "Mood detection"},
        {"name": "/split", "desc": "A/B split response"},
    ]
    return _JSON(content={"commands": cmds, "count": len(cmds)})

@router.get("/api/dashboard")
async def get_dashboard(_u=_Depends(_auth)):
    import sqlite3 as _sq
    from salmalm.constants import AUDIT_DB
    from salmalm.core import _sessions, _llm_cron, PluginLoader, SubAgent
    from salmalm.core.core import get_usage_report
    sessions_info = [{"id": s.id, "messages": len(s.messages), "last_active": s.last_active, "created": s.created}
                     for s in _sessions.values()]
    cron_jobs = _llm_cron.list_jobs() if _llm_cron else []
    plugins = [{"name": n, "tools": len(p["tools"])} for n, p in PluginLoader._plugins.items()]
    subagents = SubAgent.list_agents()
    usage = get_usage_report()
    cost_timeline = []
    try:
        _conn = _sq.connect(str(AUDIT_DB))
        cur = _conn.execute("SELECT substr(ts,1,13) as hour, COUNT(*) as cnt FROM audit_log WHERE event='tool_exec' GROUP BY hour ORDER BY hour DESC LIMIT 24")
        cost_timeline = [{"hour": r[0], "count": r[1]} for r in cur.fetchall()]
        _conn.close()
    except Exception:
        pass
    return _JSON(content={"sessions": sessions_info, "usage": usage, "cron_jobs": cron_jobs,
                          "plugins": plugins, "subagents": subagents, "cost_timeline": cost_timeline})

@router.get("/manifest.json")
async def get_manifest():
    manifest = {
        "name": "SalmAlm — Personal AI Gateway", "short_name": "SalmAlm",
        "description": "Your personal AI gateway.", "start_url": "/",
        "display": "standalone", "background_color": "#0b0d14", "theme_color": "#6366f1",
        "orientation": "any", "categories": ["productivity", "utilities"],
        "icons": [{"src": "/icon-192.svg", "sizes": "192x192", "type": "image/svg+xml", "purpose": "any"},
                  {"src": "/icon-512.svg", "sizes": "512x512", "type": "image/svg+xml", "purpose": "any maskable"}],
    }
    return _JSON(content=manifest)

@router.get("/api/search")
async def get_search(q: str = _Query(""), limit: int = _Query(20), _u=_Depends(_auth)):
    if not q:
        return _JSON(content={"error": "Missing q parameter"}, status_code=400)
    from salmalm.core import search_messages
    results = search_messages(q, limit=limit)
    return _JSON(content={"query": q, "results": results, "count": len(results)})

@router.get("/api/rag/search")
async def get_rag_search(q: str = _Query(""), n: int = _Query(5), _u=_Depends(_auth)):
    if not q:
        return _JSON(content={"error": "Missing q parameter"}, status_code=400)
    from salmalm.features.rag import rag_engine
    results = rag_engine.search(q, max_results=n)
    return _JSON(content={"query": q, "results": results})

@router.get("/api/memory/read")
async def get_memory_read(file: str = _Query(""), _u=_Depends(_auth)):
    from pathlib import PurePosixPath
    from salmalm.constants import BASE_DIR
    if not file or ".." in file:
        return _JSON(content={"error": "Invalid path"}, status_code=400)
    if PurePosixPath(file).is_absolute() or "\\" in file:
        return _JSON(content={"error": "Invalid path"}, status_code=400)
    full = (BASE_DIR / file).resolve()
    if not full.is_relative_to(BASE_DIR.resolve()):
        return _JSON(content={"error": "Path outside allowed directory"}, status_code=403)
    if not full.exists() or not full.is_file():
        return _JSON(content={"error": "File not found"}, status_code=404)
    try:
        content = full.read_text(encoding="utf-8")[:50000]
        return _JSON(content={"file": file, "content": content, "size": full.stat().st_size})
    except Exception as e:
        log.warning("[FILE] Error: %s", e)
    return _JSON(content={"error": "Internal error — check server logs"}, status_code=500)

@router.get("/api/thoughts/search")
async def get_thoughts_search(q: str = _Query(""), _u=_Depends(_auth)):
    if not q:
        return _JSON(content={"error": "query required"}, status_code=400)
    from salmalm.features.thoughts import thought_stream
    return _JSON(content={"thoughts": thought_stream.search(q)})

@router.post("/api/bookmarks")
async def post_bookmarks(request: _Request, _u=_Depends(_auth)):
    from salmalm.features.edge_cases import bookmark_manager
    body = await request.json()
    action = body.get("action", "add")
    session_id = body.get("session_id", "")
    message_index = body.get("message_index")
    if not session_id or message_index is None:
        return _JSON(content={"error": "Missing session_id or message_index"}, status_code=400)
    if action == "add":
        ok = bookmark_manager.add(session_id, int(message_index), content_preview=body.get("preview", ""),
                                  note=body.get("note", ""), role=body.get("role", "assistant"))
        return _JSON(content={"ok": ok})
    elif action == "remove":
        ok = bookmark_manager.remove(session_id, int(message_index))
        return _JSON(content={"ok": ok})
    return _JSON(content={"error": "Unknown action"}, status_code=400)

@router.post("/api/thoughts")
async def post_thoughts(request: _Request, _u=_Depends(_auth)):
    from salmalm.features.thoughts import thought_stream
    body = await request.json()
    content = body.get("content", "").strip()
    if not content:
        return _JSON(content={"error": "content required"}, status_code=400)
    tid = thought_stream.add(content, mood=body.get("mood", "neutral"))
    return _JSON(content={"ok": True, "id": tid})

@router.post("/api/thoughts/search")
async def post_thoughts_search(request: _Request, _u=_Depends(_auth)):
    from salmalm.features.thoughts import thought_stream
    body = await request.json()
    q = body.get("q", body.get("query", ""))
    if not q:
        return _JSON(content={"error": "query required"}, status_code=400)
    return _JSON(content={"thoughts": thought_stream.search(q)})
